--*****************************************************************************
--*  Copyright (c) 2017 by Michael Fischer (www.emb4fun.de) 
--*  All rights reserved.
--*
--*  Redistribution and use in source and binary forms, with or without 
--*  modification, are permitted provided that the following conditions 
--*  are met:
--*  
--*  1. Redistributions of source code must retain the above copyright 
--*     notice, this list of conditions and the following disclaimer.
--*  2. Redistributions in binary form must reproduce the above copyright
--*     notice, this list of conditions and the following disclaimer in the 
--*     documentation and/or other materials provided with the distribution.
--*  3. Neither the name of the author nor the names of its contributors may 
--*     be used to endorse or promote products derived from this software 
--*     without specific prior written permission.
--*
--*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
--*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
--*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
--*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
--*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
--*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
--*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
--*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
--*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
--*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
--*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
--*  SUCH DAMAGE.
--*
--*****************************************************************************
--*  History:
--*
--*  03.06.2017  mifi  First version.
--*****************************************************************************

--*****************************************************************************
--*  DEFINE: Library                                                          *
--*****************************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;


--*****************************************************************************
--*  DEFINE: Entity                                                           *
--*****************************************************************************

entity timer_reg is
   port (
      -- Avalon / Wishbone
      clk_i       : in  std_logic                     := '0';
      reset_i     : in  std_logic                     := '0';

      address_i   : in  std_logic_vector(2 downto 0)  := (others => '0');
      read_i      : in  std_logic                     := '0';
      readdata_o  : out std_logic_vector(15 downto 0);
      write_i     : in  std_logic                     := '0';
      writedata_i : in  std_logic_vector(15 downto 0) := (others => '0');

      -- Wishbone
      wb_ack_o    : out std_logic;

      -- None Avalon / Wishbone
      zero_o      : out std_logic;
      int_o       : out std_logic
   );
end entity timer_reg;

--*****************************************************************************
--*  DEFINE: Architecture                                                     *
--*****************************************************************************

architecture syn of timer_reg is

   --------------------------------------------------------
   -- Define all constants here
   --------------------------------------------------------
   
   constant ADR_STATUS   : std_logic_vector(2 downto 0) := "000"; -- Read / Write
   constant ADR_CONTROL  : std_logic_vector(2 downto 0) := "001"; -- Read / Write
   constant ADR_PERIOD_L : std_logic_vector(2 downto 0) := "010"; -- Write only
   constant ADR_PERIOD_H : std_logic_vector(2 downto 0) := "011"; -- Write only

   --------------------------------------------------------
   -- Define all components which are included here
   --------------------------------------------------------
   
   --------------------------------------------------------
   -- Define all local signals here
   --------------------------------------------------------

   signal readdata   : std_logic_vector(15 downto 0) := (others => '0');
   signal read_mux   : std_logic_vector(15 downto 0) := (others => '0');
   
   signal read_ack   : std_logic := '0';
   signal write_ack  : std_logic := '0';
   
   signal status     : std_logic_vector(15 downto 0) := (others => '0');  
   signal control    : std_logic_vector(15 downto 0) := (others => '0');  
   signal reload     : std_logic_vector(31 downto 0) := (others => '0');
   signal counter    : std_logic_vector(31 downto 0) := (others => '0');
   signal timer_en   : std_logic := '0';
   signal timer_zero : std_logic := '0';
   signal timer_int  : std_logic := '0';
   
begin

   --
   -- Register File
   --
   process (clk_i, reset_i)
   begin
      if (reset_i = '1') then

         control   <= (others => '0');
         write_ack <= '0';
      
      elsif rising_edge(clk_i) then
      
         -- Default
         write_ack <= '0';
      
         if (write_i = '1') then
         
            write_ack <= '1';    -- Signal normal Wishbone bus cycle termination
         
            -- Address decoder
            case address_i is
            
               when ADR_CONTROL =>
                  control <= writedata_i;
            
               when ADR_PERIOD_L =>
                  reload(15 downto 0) <= writedata_i;

               when ADR_PERIOD_H =>
                  reload(31 downto 16) <= writedata_i;
                  
               when others =>
                  -- Do nothing
            
            end case;
         end if;
         
      end if; 
   end process;
   
   
   --
   -- Timer
   --
   timer_en <= control(0);

   process (clk_i, reset_i)
   begin
      if rising_edge(clk_i) then
      
         if (timer_en = '1') then
            if (counter = x"00000000") then
               counter    <= reload;
               timer_zero <= NOT timer_zero;
               timer_int  <= '1';
            else
               counter <= std_logic_vector(unsigned(counter) - 1);
            end if;
         else
            timer_zero <= '0';   
            timer_int  <= '0';
         end if;
         
         -- Clear timer_int on STATUS write
         if ((write_i = '1') and (address_i = ADR_STATUS)) then
            timer_int <= '0';
         end if;
         
      end if;
   end process;      
   
   
   --
   -- Read Multiplexer
   --
   status(0) <= timer_int;
   
   read_mux <= status               when (ADR_STATUS   = address_i) else 
               control              when (ADR_CONTROL  = address_i) else
               reload(15 downto 0)  when (ADR_PERIOD_L = address_i) else
               reload(31 downto 16) when (ADR_PERIOD_H = address_i) else (others => '0');
               

   process (clk_i)
   begin
      if rising_edge(clk_i) then
         -- Default
         read_ack <= '0';        
      
         if (read_i = '1') then
            readdata <= read_mux;
            read_ack <= '1';     -- Signal normal Wishbone bus cycle termination
         end if;
      end if;
   end process;      

   
   --------------------------------------------------------
   -- Output
   --------------------------------------------------------

   -- Wishbone
   wb_ack_o   <= read_ack or write_ack;

   -- Wishbone / Avalon
   readdata_o <= readdata;

   -- None Wishbone / Avalon
   zero_o <= timer_zero;
   int_o  <= timer_int;

end architecture syn;

-- *** EOF ***
