--*****************************************************************************
--*  Copyright (c) 2015 by Michael Fischer (www.emb4fun.de)
--*  All rights reserved.
--*
--*  Redistribution and use in source and binary forms, with or without 
--*  modification, are permitted provided that the following conditions 
--*  are met:
--*  
--*  1. Redistributions of source code must retain the above copyright 
--*     notice, this list of conditions and the following disclaimer.
--*  2. Redistributions in binary form must reproduce the above copyright
--*     notice, this list of conditions and the following disclaimer in the 
--*     documentation and/or other materials provided with the distribution.
--*  3. Neither the name of the author nor the names of its contributors may 
--*     be used to endorse or promote products derived from this software 
--*     without specific prior written permission.
--*
--*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
--*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
--*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
--*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
--*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
--*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
--*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
--*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
--*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
--*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
--*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
--*  SUCH DAMAGE.
--*
--*****************************************************************************
--*  History:
--*
--*  20.07.2015  mifi  First version, based on my Wishbone Master.
--*  23.07.2015  mifi  Make it more universal.
--*****************************************************************************

--*****************************************************************************
--*  DEFINE: Library                                                          *
--*****************************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;


--*****************************************************************************
--*  DEFINE: Entity                                                           *
--*****************************************************************************

--
-- CPU <-> Wishbone Master
--
-- CPU_XX are the original not synchronized bus signals. 
-- The synchronisation will be done by the module here.
--
-- In case of the CPU it could be a GPMC (TI) or FSMC (STM) bus.
--
entity cpu_av_master is
   generic (
      ADDRESS_WIDTH : integer := 11
   );
   port (  
      clk_i             : in    std_logic := '0';
      rst_i             : in    std_logic := '0';

      -- CPU_A is a word address
      CPU_A             : in    std_logic_vector(ADDRESS_WIDTH-1 downto 0) := (others => '0');
      CPU_D             : inout std_logic_vector(15 downto 0);
      CPU_CS_n          : in    std_logic := '1';
      CPU_OE_n          : in    std_logic := '1';
      CPU_WE_n          : in    std_logic := '1';
      CPU_WAIT_n        : out   std_logic;

      -- Avalon Master
      avm_cs_o          : out   std_logic;
      avm_read_o        : out   std_logic;
      avm_write_o       : out   std_logic;
      avm_waitrequest_i : in    std_logic := '0';
      
      -- The Avalon address is a byte address, therefore one more byte is needed than CPU_A
      avm_address_o     : out   std_logic_vector(ADDRESS_WIDTH downto 0);
      avm_writedata_o   : out   std_logic_vector(15 downto 0);
      avm_readdata_i    : in    std_logic_vector(15 downto 0) := (others => '0')
   );
end entity cpu_av_master;


--*****************************************************************************
--*  DEFINE: Architecture                                                     *
--*****************************************************************************

architecture syn of cpu_av_master is

   --------------------------------------------------------
   -- Define all constants here
   --------------------------------------------------------

   --------------------------------------------------------
   -- Define all components which are included here
   --------------------------------------------------------
   
   --------------------------------------------------------
   -- Define all local signals here
   --------------------------------------------------------
   
   --
   -- Compared to the Wishbone bus the following control signals
   -- must be replaced by the Avalon signals:
   --
   -- wb_stb   : av_cs
   -- wb_we    : av_read / av_write
   -- wbm_ack_i: avm_waitrequest_i
   --
   
   --
   -- Control signal defines
   --   
   signal cs_s1   : std_logic := '0';  
   signal cs_s2   : std_logic := '0';  
   signal cs      : std_logic := '0';  

   signal oe_s1   : std_logic := '0';
   signal oe_s2   : std_logic := '0';
   signal oe      : std_logic := '0';
   
   signal we_s1   : std_logic := '0';
   signal we_s2   : std_logic := '0';
   signal we      : std_logic := '0';
   
   --
   -- FSM defines
   --
   type   fsm_state_t is (STATE_IDLE, STATE_READ, STATE_WRITE, STATE_DONE);
   signal fsm_state   : fsm_state_t := STATE_IDLE;

   signal wait_cnt    : std_logic_vector(5 downto 0) := (others => '0');  

   signal av_cs       : std_logic := '0';
   signal av_read     : std_logic := '0';
   signal av_write    : std_logic := '0';
   signal av_readdata : std_logic_vector(15 downto 0) := (others => '0');
   
begin

   --
   -- Synchronize CPU control signals
   --
   process (clk_i, rst_i)
   begin
      if (rst_i = '1') then
         cs_s1 <= '0';
         cs_s2 <= '0';
         
         oe_s1 <= '0';
         oe_s2 <= '0';
         
         we_s1 <= '0';
         we_s2 <= '0';
      elsif rising_edge(clk_i) then
         cs_s1 <= not CPU_CS_n;
         cs_s2 <= cs_s1;
         
         oe_s1 <= not CPU_OE_n;
         oe_s2 <= oe_s1;
         
         we_s1 <= not CPU_WE_n;
         we_s2 <= we_s1;
      end if; 
   end process;
   
   cs <= cs_s2; 
   oe <= oe_s2;
   we <= we_s2;
   
   
   --
   -- Single Process FSM
   -- Used to support the following master signals:
   --
   --    avm_cs_o             
   --    avm_read_o       
   --    avm_write_o      
   --    avm_waitrequest_i
   --
   process (clk_i, rst_i)     
   begin
      if (rst_i = '1') then
         fsm_state <= STATE_IDLE;
         av_cs    <= '0';
         av_read  <= '0';
         av_write <= '0';
         
      elsif rising_edge(clk_i) then
                     
         -- Default
         av_cs    <= av_cs; 
         av_read  <= av_read;
         av_write <= av_write;
         
         case (fsm_state) is
         
            --
            -- STATE_IDLE
            --
            when STATE_IDLE =>
               av_cs    <= '0';
               av_read  <= '0';
               av_write <= '0';
               wait_cnt <= (others =>'0');
               
               if (cs = '1') then                  -- Check for active CPU bus cycle
                  av_cs <= '1';                    -- Signal valid data transfer cycle on the Avalon bus
                  if (oe = '1') then               -- Check for CPU read access
                     fsm_state <= STATE_READ;
                     av_read   <= '1';             -- Signal a read cycle on the Avalon bus
                  end if;
                  if (we = '1') then               -- Check for CPU write access
                     fsm_state <= STATE_WRITE;
                     av_write  <= '1';             -- Signal a write cycle on the Avalon bus
                  end if;
               end if;

            --
            -- STATE_READ
            --
            when STATE_READ =>               
               if (avm_waitrequest_i = '0') then   -- Wait for a normal termination of a Avalon bus cycle                  
                  fsm_state   <= STATE_DONE;
                  av_cs       <= '0';              -- Signal end of data transfer on the Avalon bus
                  av_read     <= '0';              -- Signal end of read cycle on the Avalon bus
                  av_readdata <= avm_readdata_i;   -- Avalon data can be stored
               else               
                  fsm_state <= STATE_READ;
                  wait_cnt  <= std_logic_vector(unsigned(wait_cnt) + 1);
               end if;

            --
            -- STATE_WRITE
            --
            when STATE_WRITE =>                    
               if (avm_waitrequest_i = '0') then   -- Wait for a normal termination of a Avalon bus cycle
                  fsm_state <= STATE_DONE;
                  av_cs     <= '0';                -- Signal end of data transfer on the Avalon bus
                  av_write  <= '0';                -- Signal end of write cycle on the Avalon bus
               else
                  fsm_state <= STATE_WRITE;
                  wait_cnt  <= std_logic_vector(unsigned(wait_cnt) + 1);
               end if;

            --
            -- STATE_DONE
            --
            when STATE_DONE =>
               if ((oe = '0') and (we = '0')) then -- Wait for the end of the CPU bus cycle 
                  fsm_state <= STATE_IDLE;
               else
                  fsm_state <= STATE_DONE;
               end if;
               
            when others =>
               fsm_state <= STATE_IDLE;
               
         end case;
         
         --
         -- Abort the cycle if the bus is occupied too long.
         --
         if (wait_cnt = "111111") then
            fsm_state <= STATE_IDLE;
         end if;
         
      end if;
   end process;
   
   
   --------------------------------------------------------
   -- Output
   --------------------------------------------------------
   
   -- Control
   avm_cs_o        <= av_cs;
   avm_read_o      <= av_read;
   avm_write_o     <= av_write;
   
   -- Address 
   avm_address_o   <= CPU_A & '0';  -- The Avalon address is a byte address, 
                                    -- therefore one more byte is needed than CPU_A.
   
   -- Data
   CPU_D           <= av_readdata when (CPU_CS_n = '0' AND CPU_OE_n = '0') else (others => 'Z');
   avm_writedata_o <= CPU_D;
   
   -- Wait
   CPU_WAIT_n      <= '0' when ((av_cs = '1' AND avm_waitrequest_i = '1') OR 
                                (av_read = '1') OR (av_write= '1')) else '1';

end architecture syn;

-- *** EOF ***

