--*****************************************************************************
--*  Copyright (c) 2015-2017 by Michael Fischer (www.emb4fun.de) 
--*  All rights reserved.
--*
--*  Redistribution and use in source and binary forms, with or without 
--*  modification, are permitted provided that the following conditions 
--*  are met:
--*  
--*  1. Redistributions of source code must retain the above copyright 
--*     notice, this list of conditions and the following disclaimer.
--*  2. Redistributions in binary form must reproduce the above copyright
--*     notice, this list of conditions and the following disclaimer in the 
--*     documentation and/or other materials provided with the distribution.
--*  3. Neither the name of the author nor the names of its contributors may 
--*     be used to endorse or promote products derived from this software 
--*     without specific prior written permission.
--*
--*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
--*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
--*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
--*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
--*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
--*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
--*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
--*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
--*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
--*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
--*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
--*  SUCH DAMAGE.
--*
--*****************************************************************************
--*  History:
--*
--*  07.11.2015  mifi  First version.
--*  03.06.2016  mifi  Change SysTick from CPU ticker to FPGA timer.
--*****************************************************************************

--*****************************************************************************
--*  DEFINE: Library                                                          *
--*****************************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;


--*****************************************************************************
--*  DEFINE: Entity                                                           *
--*****************************************************************************

entity top is
   port ( 
      -- Input clock 
      CLOCK_50     : in  std_logic;

      -- GPMC bus
      GPMC_A       : in    std_logic_vector(10 downto 0);
      GPMC_D       : inout std_logic_vector(15 downto 0);
      GPMC_CS_n    : in    std_logic;
      GPMC_REN_n   : in    std_logic;
      GPMC_WEN_n   : in    std_logic;
      GPMC_RESET_n : in    std_logic;
      GPMC_WAIT_n  : out   std_logic;
      GPMC_INT_n   : out   std_logic;
      
      -- LEDs
      LED          : out std_logic_vector(7 downto 0);
      
      timer_o      : out   std_logic;
      
      -- Logic Analyzer
      la_av_cs     : out   std_logic;
      la_av_read   : out   std_logic;
      la_av_write  : out   std_logic;
      la_av_wait   : out   std_logic
   );
end entity top;


--*****************************************************************************
--*  DEFINE: Architecture                                                     *
--*****************************************************************************

architecture syn of top is

   --------------------------------------------------------
   -- Define all constants here
   --------------------------------------------------------

   --------------------------------------------------------
   -- Define all components which are included here
   --------------------------------------------------------

   component pll_sys
      port ( 
         inclk0 : in  std_logic := '0';
         c0     : out std_logic;
         c1     : out std_logic;
         locked : out std_logic 
      );
   end component pll_sys;


   component heartbeat is
      port (  
         clk_i     : in  std_logic := '0';
         counter_o : out std_logic
      );
   end component heartbeat;
   
   
   component cpu_av_master is
      generic (
         ADDRESS_WIDTH : integer
      );
      port (  
         clk_i             : in    std_logic := '0';
         rst_i             : in    std_logic := '0';
      
         -- CPU_A is a word address
         CPU_A             : in    std_logic_vector(ADDRESS_WIDTH-1 downto 0) := (others => '0');
         CPU_D             : inout std_logic_vector(15 downto 0);
         CPU_CS_n          : in    std_logic := '1';
         CPU_OE_n          : in    std_logic := '1';
         CPU_WE_n          : in    std_logic := '1';
         CPU_WAIT_n        : out   std_logic;
      
         -- Avalon Master
         avm_cs_o          : out   std_logic;
         avm_read_o        : out   std_logic;
         avm_write_o       : out   std_logic;
         avm_waitrequest_i : in    std_logic := '0';
      
         -- The Avalon address is a byte address, therefore one more byte is needed than CPU_A
         avm_address_o     : out   std_logic_vector(ADDRESS_WIDTH downto 0);
         avm_writedata_o   : out   std_logic_vector(15 downto 0);
         avm_readdata_i    : in    std_logic_vector(15 downto 0) := (others => '0')
      );
   end component cpu_av_master;


   component qsys_core is
      port (
         clk_clk           : in  std_logic                     := 'X';             -- clk
         reset_reset_n     : in  std_logic                     := 'X';             -- reset_n
         avm_cs_i          : in  std_logic                     := 'X';             -- cs_i
         avm_address_i     : in  std_logic_vector(11 downto 0) := (others => 'X'); -- address_i
         avm_read_i        : in  std_logic                     := 'X';             -- read_i
         avm_waitrequest_o : out std_logic;                                        -- waitrequest_o
         avm_readdata_o    : out std_logic_vector(15 downto 0);                    -- readdata_o
         avm_write_i       : in  std_logic                     := 'X';             -- write_i
         avm_writedata_i   : in  std_logic_vector(15 downto 0) := (others => 'X'); -- writedata_i
         gpio0_o           : out std_logic_vector(15 downto 0);                    -- o
         gpio0_i           : in  std_logic_vector(15 downto 0) := (others => 'X'); -- i
         gpio1_o           : out std_logic_vector(15 downto 0);                    -- o
         gpio1_i           : in  std_logic_vector(15 downto 0) := (others => 'X'); -- i
         timer0_zero_o     : out std_logic;                                        -- zero_o
         timer0_int_o      : out std_logic                                         -- int_o
      );
   end component qsys_core;


   --------------------------------------------------------
   -- Define all local signals here
   --------------------------------------------------------
   
   signal sys_clk        : std_logic := '0';
   signal pll_locked     : std_logic := '0';
   signal reset          : std_logic := '0';
   signal reset_s1       : std_logic := '1';
   signal reset_s2       : std_logic := '1';
   signal reset_s3       : std_logic := '1';
   signal sys_rst        : std_logic;
   signal clk_10         : std_logic;
   
   signal av_cs          : std_logic;
   signal av_address     : std_logic_vector(11 downto 0);
   signal av_read        : std_logic;
   signal av_waitrequest : std_logic;
   signal av_readdata    : std_logic_vector(15 downto 0);
   signal av_write       : std_logic;
   signal av_writedata   : std_logic_vector(15 downto 0);
   
   signal gpio           : std_logic_vector(15 downto 0);
   
   signal timer_int      : std_logic;

begin

   --
   -- PLL
   --
   inst_pll_sys : pll_sys
      port map ( 
         inclk0 => CLOCK_50,
         c0     => sys_clk,
         c1     => clk_10,
         locked => pll_locked
      );
      
      
   --
   -- In general it is a bad idea to use an asynchhronous Reset signal.
   -- But it is only a bad idea in case of asynchhronous deasserting. 
   -- Therefore the deasserting of the Reset signal must be synchronized.
   --               
                              
   -- Asynchronous assert  
   reset <= '1' when ((GPMC_RESET_n = '0') OR (pll_locked = '0')) else '0';

   -- Synchronize deassert
   process (sys_clk, reset)
   begin
      if (reset = '1') then
         reset_s1 <= '1';
         reset_s2 <= '1';
         reset_s3 <= '1';
      elsif rising_edge(sys_clk) then
         reset_s1 <= '0';
         reset_s2 <= reset_s1;
         reset_s3 <= reset_s2; 
      end if;
   end process;   

   -- The deassert edge is now synchronized   
   sys_rst <= reset_s3; 
      
   -- Heartbeat
   inst_heartbeat : heartbeat 
      port map ( 
         clk_i     => clk_10,
         counter_o => LED(7)
      );

      
   -- GPMC to Avalon Master
   inst_gpmc_av_master : cpu_av_master   
      generic map (
         ADDRESS_WIDTH     => GPMC_A'length
      )
      port map ( 
         clk_i             => sys_clk,
         rst_i             => sys_rst,
      
         -- None Avalon
         CPU_A             => GPMC_A,
         CPU_D             => GPMC_D,    
         CPU_CS_n          => GPMC_CS_n,
         CPU_OE_n          => GPMC_REN_n,
         CPU_WE_n          => GPMC_WEN_n,         
         CPU_WAIT_n        => GPMC_WAIT_n,
      
         -- Avalon Master
         avm_cs_o          => av_cs,         
         avm_address_o     => av_address,
         avm_read_o        => av_read,
         avm_waitrequest_i => av_waitrequest,
         avm_readdata_i    => av_readdata,
         avm_write_o       => av_write,
         avm_writedata_o   => av_writedata
      );
      
   -- Qsys
   inst_qsys : qsys_core
      port map (
         clk_clk           => sys_clk,
         reset_reset_n     => NOT sys_rst,
         
         avm_cs_i          => av_cs,         
         avm_address_i     => av_address,
         avm_read_i        => av_read,
         avm_waitrequest_o => av_waitrequest,
         avm_readdata_o    => av_readdata,
         avm_write_i       => av_write,
         avm_writedata_i   => av_writedata,
         
         gpio0_o           => gpio,
         gpio0_i           => open,
         gpio1_o           => open,
         gpio1_i           => gpio,
         
         timer0_zero_o     => open,
         timer0_int_o      => timer_int
      );
         

   --------------------------------------------------------
   -- Output
   --------------------------------------------------------
   
   -- LED 
   LED(6 downto 0) <= gpio(6 downto 0);
   
   GPMC_INT_n <= NOT timer_int;
   timer_o    <= timer_int;     
            
   -- Logic Analyzer   
   la_av_cs    <= av_cs;
   la_av_read  <= av_read; 
   la_av_write <= av_write;
   la_av_wait  <= av_waitrequest;
         
end architecture syn;

-- *** EOF ***

